import tensorflow as tf

class CRF(tf.keras.layers.Layer):
    def __init__(self, k, n, m,**kwargs):
        """
        a CRF (Conditional Random Field) layer
        for implementation details see: https://github.com/..../...
        k    (int)----> number of tags without START , END and padding tags
        n    (int)----> number of words in the sentence (maximum length)
        m    (int)----> number of features in the word embedding
        """
        self.k = k
        self.m = m
        self.n = n
        super(CRF, self).__init__(**kwargs)
    def build(self, input_shape):
        # intialize weights randomly using uniform distribution
        # W is represent each tag as a row and each feature as a column
        # W has a shape of (m, k+2)
        # E is the transition matrix between the tags (not T to prevent confusion with the time step)
        # E has a shape of (k+2, k+2)
        W = tf.random.uniform([self.m,self.k],0,1, dtype = tf.float64)
        W = tf.concat([tf.zeros([self.m,1], dtype = tf.float64), W, tf.zeros([self.m,1], dtype = tf.float64)], axis = 1)
        E = tf.random.uniform([self.k+1,self.k+1], 0, 1, dtype = tf.float64)
        E = tf.concat([E, tf.zeros([1,self.k+1], dtype = tf.float64)], axis = 0)
        E = tf.concat([tf.zeros([self.k+2,1], dtype = tf.float64), E], axis = 1)
        self.E = self.add_weight(shape = E.shape,\
                                initializer = tf.keras.initializers.Constant(E),\
                                trainable = True,\
                                dtype = tf.float64)
        self.W = self.add_weight(shape = W.shape,\
                                initializer = tf.keras.initializers.Constant(W),\
                                trainable = True,\
                                dtype = tf.float64)
    @tf.function
    def Veterbi(self, U):
        # for more information about the impementation of the Veterbi algorithm see: https://github.com/..../...
        labels  = tf.range(1, self.k+1, dtype=tf.int32)
        pi      = tf.zeros([labels.shape[0], self.n], dtype = tf.float64)
        # Base case initialization
        pi = tf.tensor_scatter_nd_update(pi, [[i, 0] for i in range(self.k)], U[1, 1:-1])

        for t in range(2, self.n + 1):
            pi_t_minus_2 = tf.expand_dims(pi[:, t-2], axis=1)
            temp = pi_t_minus_2 + self.E[1:-1, 1:-1] + U[t, 1:-1]
            pi = tf.tensor_scatter_nd_update(pi, [[i, t-1] for i in range(self.k)], tf.reduce_max(temp, axis=0))

        # Backtracking
        best = tf.zeros((self.n + 2), dtype=tf.int32)
        best = tf.tensor_scatter_nd_update(best, [[self.n+1]], [labels[-1] + 1])
        best = tf.tensor_scatter_nd_update(best, [[self.n]], [labels[tf.argmax(pi[:, -1])]])
        for t in range(self.n - 1, 0, -1):
            last_best = best[t+1]
            temp_sum = pi[:, t-1] + self.E[1:-1, last_best] + U[t+1, last_best]
            best = tf.tensor_scatter_nd_update(best, [[t]], [labels[tf.argmax(temp_sum)]])
        return best
    @tf.function
    def Z_Forward(self, U):
        # for more information about the impementation of the Forward algorithm see: https://github.com/..../...
        # base case
        phi  = tf.reshape(tf.exp(U[1,1:-1]), [-1,1])
        for i in range(2, self.n+1):
            temp = tf.repeat(phi, phi.shape[0], axis = 1) * tf.exp(self.E[1:-1,1:-1] + U[i,1:-1])
            phi  = tf.reshape(tf.reduce_sum(temp, axis = 0), [-1,1])
        return tf.reduce_sum(phi)
    
    def call(self, data, return_loss = True):
        if return_loss:
            # unpack the data
            inputs,Y = data
            # calculate unary potentials scores over the patch
            U     = tf.cast(inputs, tf.float64) @ self.W
            Y_hat = tf.map_fn(self.Veterbi, U, dtype = tf.int32, parallel_iterations = 11)
            # calculate the loss
            emissions   = tf.map_fn(fn =\
                          lambda y:tf.stack([y[1:-2], y[2:-1]], axis = 1), elems = Y)
            unaries     = tf.map_fn(fn =\
                          lambda y:tf.stack([tf.range(1,y.shape[0]-1), y[1:-1]], axis = 1),elems = Y)
            Z           = tf.map_fn(fn = self.Z_Forward, elems = U, parallel_iterations = 11)
            F           = tf.reduce_sum(tf.gather_nd(U, unaries, batch_dims = 1), axis = 1)
            F          += tf.reduce_sum(tf.gather_nd(self.E, emissions, batch_dims = 0), axis = 1)
            loss        = tf.reduce_mean(-1* (F - tf.math.log(Z)))
            return Y_hat,loss
        else:
            inputs = data
            U      = tf.cast(inputs, tf.float64) @ self.W
            Y_hat  = tf.map_fn(self.Veterbi, U, dtype = tf.int32, parallel_iterations = 4)
            return Y_hat


import numpy as np
import tensorflow as tf
from bpemb import BPEmb
import pickle
import os
from .NER import NER
class Tagger:
    """
        Tagging class is responsible for tagging the input text (Arabic)
        with the user defined tags + the tags of two pretrained models
        this are going to help he user to build a model knowing only the 
        tags (usually single tag).

        Example
        --------
        tagger = Tagger()
        tagger.intialize_models()
        inputs = [['السلام', 'عليكم', 'كم', 'سعر', 'الخلاط'],
                 ['ما', 'هي', 'مواصفات', 'البوتجاز', 'الي', 'في', 'الصورة']]
        tags =  [['DEVICE', 'O', 'O', 'O', 'O'],
                 ['O', 'O',' O', 'DEVICE', 'O', 'O', 'O']]
        # define udt
        user_defined_tags = ['DEVICE']
        # the output is a tuple of two tuples each tuple contains the input embeddings
        #  and the tags as integers 
        # the first tuple is the input embeddings and the tags of the first model
        # the second tuple is the input embeddings and the tags of the second model
        # you can use this data to train CRF model or NER(Bi-LSTM + CRF) model
        train_data1, train_data2 = tagger.get_data(inputs, tags, user_defined_tags)
        # assume we have trained two model model1 & model2
        txt = "ما هي مواصفات الغسالة الي في الصورة"
        >>> tagger.tags1["DEVICE"]
        13
        >>> tagger.tags2["DEVICE"]
        8
        predictions = tagger.predict([model1, model2], txt.split(), [13, 8])
        # if the models are trained well the output should be
        #  array(['الغسالة'], dtype='<U7')
    """
    def __init__(self):
        pass
    def intialize_models(self):
        """
        intialize_models method is responsible for loading the pretrained models
        - BPEmb model which is responsible for embedding the input text 
        for more information about BPEmb model please visit https://github.com/bheinzerling/bpemb
        - CRF models & tags see the documentation of NER class at https://github.com/..../...
        """
        self.embedding =  BPEmb(dim = 300, lang = "arz")
        current_dir = os.path.dirname(__file__)
        self.model1   = self.load(os.path.join(current_dir, 'models', "CRF_model_1"))
        self.model2   = self.load(os.path.join(current_dir, 'models', "CRF_model_2"))
        self.tags1    = self.get_label_mapping(self.load_tags(os.path.join(current_dir, 'tags', 'tags1')))
        self.tags2    = self.get_label_mapping(self.load_tags(os.path.join(current_dir, 'tags', 'tags2')))
    def predict(self, models, words, tags_ids):
        """
        predict method is responsible for predicting the tags of the input list of words
        it makes sure to give you only the tags that are common between the models and which
        has been defined inside tags_ids list see examples for more information
        inputs:
            models (List[keras.Model])------> list of pretrained models (CRF or NER)
            words (List[str])---------------> list of words to be tagged
            tags_ids (List[int])------------> list of identical tags ids to be searched and filtered 
        outputs:
            prediction (numpy.ndarray)------> list of words with the predicted tags
        Example
        --------
        # assume we have defined Tagger object then prepared the data (see class example)
        # trained two models model1 and model2 
        >>> tagger.tags1["DEVICE"]
        13
        >>> tagger.tags2["DEVICE"]
        8
        # model 1 has assigned tag_id 8 to DEVICE and model 2 has assigned tag_id 13 to DEVICE
        # the model will only select token that has been assigned labels 8 and 13
        #  by model 1 and model 2 respectively
        txt = "ما هي مواصفات الغسالة الي في الصورة"
        predictions = tagger.predict([model1, model2], txt.split(), [13, 8])
        # if the model is trained well the output should be
        #  array(['الغسالة'], dtype='<U7')
        """
        target = 0 
        for model,tag_id in zip(models, tags_ids):
            X      = self.preprocess_input([words], model.n)
            Y_hat  = model.predict(X, verbose = False, batch_size = len(X))
            target = target + tf.cast(Y_hat == tag_id, tf.int32)
        target     = (target == len(models))[0,1:len(words)+1]
        prediction = np.array(words)[target.numpy()]
        return prediction
    def load_tags(self, path):
        # private method 
        # load tags from the file then return them
        # the tags stored as lists
        with open(f"{path}.pkl",'rb') as file:
            tags = pickle.load(file)
        return tags
    def load(self, path):
        # private method
        # load the model weights and parameters given the path
        # return the model
        # used to laod models 1 and 2
        with open(str(path)+".pkl", 'rb') as file:
            container = pickle.load(file)
        model = NER(container["units"], container["k"], container["n"], container["m"])
        model.lstm.build((None, container["n"], container["m"]))
        model.crf.build((None, container["n"], container["units"]*2))
        model.set_weights(container["model_weights"])
        print("model weights has been loaded!")
        return model
    def get_label_mapping(self, labels):
        # private method
        # map each lable to it's corresponding integer, starting from 1
        # save it to a dictionary then return it
        label_map   = dict() 
        for i,label in enumerate(labels):
            label_map[label] = i+1
        return label_map
    
    def vectorize_X(self, words):
        # private method
        # vectorize the input words using the embedding model BPEmb
        w_matrix    = []
        # start word are set of zeros
        w_matrix = [[0]*300]
        for word in words:
            embeds = self.embedding.embed(word)
            if len(embeds) > 1:
                embeds = np.average(embeds, axis = 0).reshape([1,embeds.shape[1]])
            w_matrix.append(embeds.tolist()[0])
        # end vector are set of 100s
        w_matrix.append([100]*300)
        return w_matrix
    
    def Normalize_length(self, x, padd, l):
        # private method
        # normalize the length of the input list x to be equal to l (maximum length)

        # if the length of the input list is less than maximum length append paddings
        # if the length of the input list is greater than maximum length truncate the list
        if len(x) < l:
            x.extend([padd]*(l - len(x)))
        elif len(x) > l:
            x = x[:l]
        return x
    def preprocess_input(self, words, n):
        # private method
        # preprocess the input words to be ready for the model

        # preprocessing steps are:
        # 1- normalize the length of the input list to be equal to n
        # 2- vectorize the input list using the embedding model
        # 3- return the vectorized input list as a tensor
        X = []
        for w in words:
            x = self.Normalize_length(w, "<PAD>", n)
            x = self.vectorize_X(x)
            X.append(x)
        X = tf.constant(X, dtype = tf.float64)
        return X
    def get_data(self, inputs, tag_lists, user_defiend_tags):
        """
        get_data method is responsible for preparing the data to be used in training
        the CRF model or NER model.
        inputs:
            inputs (List[List[str]])------> list of input text to be tagged "inputs can be iterable"
            tag_lists (List[List[str]])---> list of tags of the input text, the tags must be some unique tags 
            and the unknown tags are "O"  "tag_lists can be iterable"
            user_defined_tags (List[str])-> list of user defined tags whcih is unique tags in the tags lists except "O"
        outputs:
            train_data1 (Tuple)-----------> tuple of input embeddings Tensor and the tags of the first model
            train_data2 (Tuple)-----------> tuple of input embeddings Tensor and the tags of the second model
        Example
        --------
        tagger = Tagger()
        tagger.intialize_models()
        inputs = [['السلام', 'عليكم', 'كم', 'سعر', 'الخلاط'],
                 ['ما', 'هي', 'مواصفات', 'البوتجاز', 'الي', 'في', 'الصورة']]
        tags =  [['DEVICE', 'O', 'O', 'O', 'O'],
                 ['O', 'O',' O', 'DEVICE', 'O', 'O', 'O']]
        # define udt
        user_defined_tags = ['DEVICE']
        train_data1, train_data2 = tagger.get_data(inputs, tags, user_defined_tags)
        >>> print(train_data1[0].shape, train_data1[1].shape)
        (2, 42, 300) (2, 42)

        """
        self.tags1 = self.get_label_mapping(list(self.tags1.keys())+user_defiend_tags)
        self.tags2 = self.get_label_mapping(list(self.tags2.keys())+user_defiend_tags)
        pos = []
        replacements = [[], []]
        for i, tag_list in enumerate(tag_lists):
            for j, tag in enumerate(tag_list):
                if tag != "O":
                    pos.append([i,j+1])
                    replacements[0].append(self.tags1[tag])
                    replacements[1].append(self.tags2[tag])
        X_1 = self.preprocess_input(inputs, self.model1.n)
        Y_1 = self.model1.predict(X_1, verbose = False, batch_size= len(X_1))
        Y_1 = tf.concat([tf.zeros([Y_1.shape[0], Y_1.shape[1]-1]),\
                        tf.ones([Y_1.shape[0], 1])],axis = 1) + Y_1
        Y_1  = tf.tensor_scatter_nd_update(Y_1, pos, replacements[0])
        Y_1  = tf.cast(Y_1, tf.int32)
        
        X_2  = self.preprocess_input(inputs, self.model2.n)
        Y_2  = self.model2.predict(X_2, verbose = False, batch_size= len(X_2))
        Y_2  = tf.concat([tf.zeros([Y_2.shape[0], Y_2.shape[1]-1]),\
                          tf.ones([Y_2.shape[0], 1])],axis = 1) + Y_2
        Y_2  = tf.tensor_scatter_nd_update(Y_2, pos, replacements[1])
        Y_2  = tf.cast(Y_2, tf.int32)
        train_data1 = (X_1,Y_1)
        train_data2 = (X_2,Y_2)
        return train_data1, train_data2


import pickle
from .NER import NER

def save(path, model):
    """
    Save the model parameters to a file
    inputs:
        path (str)-----> the path to save the model must be the full path + name of file without any extension
        model (NER)----> the model to save
    """
    container = {"k":model.k, "units":model.units,\
                 "m":model.m,"n":model.n,\
                 "model_weights":model.get_weights()}
    with open(str(path)+".pkl",'wb') as file:
        pickle.dump(container, file)
        print("model parameters has been saved !")


def load(path):
    """
    load the model parameters from a file
    input:
        path (str)-----> the path to the model must be the full path + name of file without any extension
    output:
        model (NER)----> the model with the loaded parameters
    """

    with open(str(path)+".pkl", 'rb') as file:
        container = pickle.load(file)
    model = NER(container["units"], container["k"], container["n"], container["m"])
    model.lstm.build((None, container["n"], container["m"]))
    model.crf.build((None, container["n"], container["units"]*2))
    model.set_weights(container["model_weights"])
    print("model weights has been loaded!")
    return model


import tensorflow as tf
from .CRF import CRF
from tensorflow.keras.layers import LSTM
from tensorflow.keras.layers import Bidirectional
class NER(tf.keras.Model):
    def __init__(self, units, k, n , m, **kwargs):
        """
        a NER (Named Entity Recognizer) model that uses a BI-LSTM + CRF layers to predict the tags of the words in a sentence
        units (int)---> number of hidden units in the LSTM
        k    (int)----> number of tags without START , END and padding tags
        n    (int)----> number of words in the sentence (maximum length)
        m    (int)----> number of features in the word embedding
        udt  (List[int])----> User Defined Tags (list of tags that the model should consider in the accuracy calculation)
        """
        self.units = units   
        self.k     = k     
        self.n     = n       
        self.m     = m
        self.udt   = kwargs.pop("udt", None)
        super(NER, self).__init__(**kwargs)
        # metrics to calculate the loss and accuracy
        self.loss_tracker = tf.keras.metrics.Mean(name="loss")
        self.accuracy_tracker = tf.keras.metrics.Mean(name="accuracy")
        self.udt_accuracy_tracker = tf.keras.metrics.Mean(name="udt_accuracy")
        self.lstm = Bidirectional(LSTM(self.units, return_sequences=True))
        self.crf  = CRF(k, n, self.units * 2)
    def call(self, data, training = False):
        # if training is True, the data is a tuple of X and Y
        # if training is False, the data is a tensor X which is used to predict the Y
        if training:
            X, Y = data
            return self.crf((self.lstm(X), Y))
        else:
            X = data 
            return self.crf(self.lstm(X), return_loss = False)
        
    def train_step(self, data):
        x,y = data
        # if the shape of the input is (None, n+2, m) then we need to make it (1, n+2, m)
        if x[0].shape[0] == None:
            y   = tf.expand_dims(y[0], 0)
            x   = tf.expand_dims(x[0][0],0)
        else:
            x = x[0]
        # do forward pass
        with tf.GradientTape() as tape:
            y_hat, loss = self((x, y), training = True)
            gradients = tape.gradient(loss, self.trainable_variables)
            self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
        accu = tf.reduce_mean(tf.cast(tf.equal(y, y_hat), tf.float32))
        # Calculate the accuracy of the UDT
        # if udt is None then use all the tags
        if self.udt:
            mask = tf.reduce_any(tf.equal(tf.expand_dims(y, axis=-1), self.udt), axis=-1)
            # Calculate temp using the mask
            temp = tf.cast(y == y_hat, dtype=tf.int32) * tf.cast(mask, dtype=tf.int32)
            temp = tf.reduce_sum(temp)
            # Calculate udt_accu
            udt_accu = temp / tf.reduce_sum(tf.cast(mask, dtype=tf.int32))
        else:
            udt_accu = accu
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            elif metric.name == "accuracy":
                metric.update_state(accu)
            elif metric.name == "udt_accuracy":
                metric.update_state(udt_accu)
            else:
                metric.update_state(y, y_hat)
        return {m.name: m.result() for m in self.metrics}
    
    def test_step(self, data):
        # Unpack the data
        x,y = data
        if x[0].shape == None:
            y   = tf.expand_dims(y[0], 0)
            x   = tf.expand_dims(x[0][0],0)
        else:
            x = x[0]
        # Compute predictions
        y_hat, loss = self((x, y), training = True)
        # Update the metrics.
        accu     = tf.reduce_mean(tf.cast(tf.equal(y, y_hat), tf.float32))
        if self.udt:
            mask = tf.reduce_any(tf.equal(tf.expand_dims(y, axis=-1), self.udt), axis=-1)
            # Calculate temp using the mask
            temp = tf.cast(y == y_hat, dtype=tf.int32) * tf.cast(mask, dtype=tf.int32)
            temp = tf.reduce_sum(temp)
            # Calculate udt_accu
            udt_accu = temp / tf.reduce_sum(tf.cast(mask, dtype=tf.int32))
        else:
            udt_accu  = accu
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            elif metric.name == "accuracy":
                metric.update_state(accu)
            elif metric.name == "udt_accuracy":
                metric.update_state(udt_accu)
            else:
                metric.update_state(y, y_hat)
        return {m.name: m.result() for m in self.metrics}
    


from setuptools import setup, find_packages

setup(
    name="ArabicTagger",  
    version="0.1.4b1",
    packages=find_packages(),
    install_requires=["tensorflow==2.10.0",\
                      "keras==2.10.0",\
                       "bpemb", "numpy"],
    include_package_data=True,
    author="Eslam Tarek Farouk",
    description="A BI-LSTM & CRF model implemented in Keras & Arabic Tagger",
    url="https://github.com/yourusername/my_package",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
